# Generated by Django 4.2.2 on 2023-09-05 08:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0013_alter_annotationpreset_permission_group_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationfield",
            name="validator_regex",
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
    ]
